#!/usr/bin/perl

# ortho.pl -- Verificateur automatique de fichiers pour KDE France
#               Copyright
#		Franois-Xavier Duranceau <duranceau@kde.org> Dec 2000
# 		$Id

package Ortho;

use File::Basename;
use Getopt::Long;

use strict;
use diagnostics;

my $APP_NAME = 'ortho';

my $DEBUG = 0;
my $AUTOFIX = 0;

#my $ERRORS = 0;
#my $INPUT_LINE = 0;
#my $INPUT_FILE = '';


############################## COMMON SUBROUTINES #################################

sub debug{
  if( $DEBUG ){
    my @a = caller(1);
    my ($pack,$file,$line,$sub,$hargs,undef,$eval,$require) = @a;

    print STDERR "\n";
    while( my $arg = shift ){
      print STDERR "debug ($sub:$line) ", $arg, "\n";
    }
  }
}

sub suicide{
  my @a = caller(1);
  my ($pack,$file,$line,$sub,$hargs,undef,$eval,$require) = @a;

  print STDERR "\n";
  while( my $arg = shift ){
    print STDERR 'suicide ($sub:$line): ';
    print STDERR $arg, "\n";
  }
  exit -1;
}

sub warning{
  my @a = caller(1);
  my ($pack,$file,$line,$sub,$hargs,undef,$eval,$require) = @a;

  print STDERR "\n";
  while( my $arg = shift ){
    print STDERR 'warning ($sub:$line): ';
    print STDERR $arg, "\n";
  }
}

sub set_debug($){
	$DEBUG = shift;
}

############################### RULESET MANAGEMENT ################################

my $COMMON_RULESET_NAME = 'common';
my $RULESET_EXTENSION = 'rules';

my $RULESET_FILENAME = '';
my $RULESET_LINE = 0;

my %RULESET_DATA = ();
my @RULESET_MANDATORY_KEYS = ('hint');

sub convert_entities{
  my $str = shift;
  
  $str =~ s/&amp;/&/g;
  $str =~ s/&cr;/\r/g;
  $str =~ s/&lf;/\n/g;
  $str =~ s/&lt;/</g;
  $str =~ s/&gt;/>/g;
  $str =~ s/&sp;/ /g;
  $str =~ s/&quot;/\"/g;
  
  return $str;
}

sub load_rulesets($){
  %RULESET_DATA = ();

  #load the common rules
  load_one_ruleset( $COMMON_RULESET_NAME );
  
  #load the specific rules (can override the common ruleset)
  load_one_ruleset( shift );  
}

sub ruleset_read_one_line{
  my $l;
				   
  while( defined( $l = <RULESET_FH> ) ){
    chop $l;
    $RULESET_LINE++;
    debug "ruleset_read_one_line: line=`$l' ($RULESET_FILENAME:$RULESET_LINE)";
    next if( $l =~ m/^\s*\#/ );
    next if( $l =~ m/^\s*$/ );
    return $l;
  }    
  return undef;
}

sub ruleset_parse_line($){
  my $l = shift;
  debug "+ruleset_parse_line (l = `$l')";
  
  $l =~ s/^\s+//;
  $l =~ s/\s+$//;

  if( $l =~ m/^(\w+)=\"(.*?)\"$/ ){
    return ($1, $2);

  }elsif( $l =~ m/^(\w+)/ ){
    my $tok = $1;
    my $it = {};

    while( $l =~ m/(\w+)=\"(.*?)\"\s*/g ){
      debug "ruleset_parse_line: parsed (key,val)=(`$1',`$2')";

      suicide "syntax error: $1 redefined in $l at $RULESET_FILENAME:$RULESET_LINE"
	if( defined $it->{$1} );

      $it->{$1} = convert_entities( $2 );
    }
    suicide "ruleset_parse_line: unable to parse `$l' at $RULESET_FILENAME:$RULESET_LINE"
      if( $l !~ m/\s*$/g );

    return ($tok, $it);
  }
}

sub ruleset_check_args{
  my $key = shift;
  my $arg = shift;
  debug "+ruleset_check_args (key = $key)";

  my %allowed = ();
  while( defined( my $l = shift ) ){
    $allowed{$l} = 1;
  }

  foreach my $k (keys %{$arg}){
    suicide "`$k' is not an expected argument for `$key' ($RULESET_FILENAME:$RULESET_LINE)"
      if( ! defined $allowed{$k} );
  }
  debug "-ruleset_check_args";
}

sub ruleset_check_rule($$){
  my ($it, $name) = @_;

  # check if all mandatory fields are filled
  foreach my $k (@RULESET_MANDATORY_KEYS){
    suicide "rule check failed (missing `$k' definition for `$name')"
      if( ! defined $it->{$k} );
  }
}

sub ruleset_check_validation_id($){
  my $it = shift;

  my $id = $it->{id};

  foreach my $k (keys %RULESET_DATA){
    my $rule = $RULESET_DATA{$k};
    foreach my $validator ( @{ $rule->{validators} } ){
      if( $validator->{id} eq $id ){
	my $after = $validator->{after};
	my $before = $validator->{before};

	my $msg = "id $id is already defined for the following rule:\n";
	$msg .=   "  validation of `$k' - before = `$before', after = `$after'";
	suicide $msg;
      }
    }
  }
}

sub load_one_ruleset($){
  my $mode = shift;
  debug "+load_one_ruleset (mode = `$mode')";

  $RULESET_FILENAME = (dirname $0) . "/$APP_NAME" . '-' . $mode . ".$RULESET_EXTENSION";
  print STDERR 'Loading ruleset ', (basename $RULESET_FILENAME);

  debug "load_one_ruleset: filename = `$RULESET_FILENAME'";

  open RULESET_FH, $RULESET_FILENAME or suicide "can't find ruleset `$RULESET_FILENAME' for $mode";
  $RULESET_LINE = 0;

  my $l;
  my $item = {};
  my $about = '';

  while( defined( $l = ruleset_read_one_line() ) ){
    if( $l =~ m/^\s*\[(.+)\]/ ){
      debug "load_one_ruleset: new rule";
      
      if( $about ne '' ){ # avoid storing empty data when reading the 1st rule
	ruleset_check_rule( $item, $about ) if( $about ne '' );
	debug "storing data about `$about'";

	$RULESET_DATA{$about} = $item; # store the previous rule (it's correct !)
	print STDERR '.'; # progress feedback
      }

      $about = convert_entities( $1 );
      suicide "syntax error: missing or empty [???] section"
	if( $about eq '' );

      if( defined $RULESET_DATA{$about} ){
	warning "previous declaration of $about is overriden at $RULESET_FILENAME:$RULESET_LINE";
	undef $RULESET_DATA{$about};
      }

      $item = {};

    }else{
      debug "load_one_ruleset: new rule description line";

      suicide "syntax error: missing [???] section"
	if( $about eq '' );

      my ($key, $arg) = ruleset_parse_line( $l );

      if( $key eq 'hint' ){
	$item->{hint} = $arg;
	debug "load_one_ruleset: hint defined for `$about'";

      }elsif( $key eq 'valid' ){
	ruleset_check_args( $key, $arg, 'before', 'after', 'id', 'ctx' );
	$arg->{before} = '' if( ! defined $arg->{before} );
	$arg->{after} = '' if( ! defined $arg->{after} );
	$arg->{ctx} = '' if( ! defined $arg->{ctx} );

	suicide "`id' is not defined for the validation rule at $RULESET_FILENAME:$RULESET_LINE"
	  if( ! defined $arg->{id} );

	suicide "valid rule without arg at $RULESET_FILENAME:$RULESET_LINE"
	  if( ! defined $arg );
	push @{$item->{validators}}, $arg;
	debug "load_one_ruleset: valid defined for `$about'";

      }elsif( $key eq 'autofix' ){
	ruleset_check_args( $key, $arg, 'before', 'after' );
	$arg->{before} = '' if( ! defined $arg->{before} );
	$arg->{after} = '' if( ! defined $arg->{after} );

	suicide "autofix rule without arg at $RULESET_FILENAME:$RULESET_LINE"
	  if( ! defined $arg );
	$item->{autofix} = $arg;
	debug "load_one_ruleset: autofix defined for `$about'";

      }else{
	suicide "syntax error: unexpected `$key' at `$RULESET_FILENAME:$RULESET_LINE'";
      }
    }
  }
  if( $about ne '' ){ # avoid storing empty data when reading the 1st rule
    ruleset_check_rule( $item, $about ) if( $about ne '' );
    debug "storing data about `$about'";
      
    $RULESET_DATA{$about} = $item; # store the previous rule (it's correct !)
    print STDERR '.'; # progress feedback
  }
  close RULESET_FH;

  print STDERR "ok\n"; # progress feedback

  ruleset_print_rules() if( $DEBUG );
  debug "-load_one_ruleset";
}

sub ruleset_print_rules{
  print "Ruleset\n";
  print "=======\n\n";
  
  foreach my $tok ( keys %RULESET_DATA ){
    print "token: `$tok'\n";
    my $rule = $RULESET_DATA{$tok};
    print "  hint    : `$rule->{hint}'\n";
    print "  a_fix  : `$rule->{autofix}{after}'\n";
    print "  b_fix  : `$rule->{autofix}{before}'\n";
#    print "  t_fix  : `$rule->{token_fix}'\n";
    print "  Validators :\n";
    foreach my $validator ( @{ $rule->{validators} } ){
      print "  after  : `$validator->{after}'\n";
      print "  before : `$validator->{before}'\n";
      print "  id     : `$validator->{id}'\n\n";
      print "  ctx    : `$validator->{ctx}'\n\n";
    }
  }
}

############################## Generic processing ##############################

sub check_text($$){
	# Returns
	#  (0, "error message") if it fails
	#  (1, undef) if it succeeds

	my ($string, $ctx) = @_;
	my $ok = 1;
	my @causes = ();

	debug "+check_string: string = `$string'";

	foreach my $tok ( keys %RULESET_DATA ){
		debug "check_string: tok is `$tok'";
		
		pos $string = 0; # restart from beginning of $string
		
		while( $string =~ /$tok/gi ){
			debug "check_string: found `$tok'";
	  
			my $valid = 0;
			my $rule = $RULESET_DATA{$tok};
			
			foreach my $validator ( @{ $rule->{validators} } ){
		  
				debug "checking rule $validator->{id}";
				
				if( $validator->{ctx} ne '' ){
					next if( $validator->{ctx} ne $ctx );
				}
				
				if( defined($validator->{after}) ){ # test the left part
					my $l = $validator->{after};  debug "?after `$l'?";
					
					if ( $` =~ /$l$/ ){
						$valid = 1;  debug "-> ok";
					}else{
						$valid = 0;  debug "-> error";
						next; # no need to continue as this validation rule already failed
					}
				}
				
				if( defined($validator->{before}) ){ # test the right part
					my $r = $validator->{before};  debug "?before `$r'?";
					
					if ( $' =~ /^$r/ ){
						$valid = 1;  debug "-> ok";
					}else{
						$valid = 0;  debug "-> error";
					}
				}
				if( $valid ){
					debug "$validator->{id} validates the current token";
					last;
				}
			}
			
			if( !$valid ){
				my $where = pos $string;
				
				my $cause = "bad pattern `$tok' (offset: $where)";
			    $cause .= " in context: `$ctx'" if( $ctx ne '' );
				$cause .= "\nHint: $rule->{hint}";
				push @causes, $cause;
				$ok = 0;
#				return (0, $cause);
			}
		}
	}
	return ($ok, @causes);
}



sub fix_string($){
  my $str = shift;
  my $valid = 0;
  
  foreach my $tok (keys %RULESET_DATA){
    debug("fix_string: `$tok' in `$str'");

    my $rule = $RULESET_DATA{$tok};

    #        next if( not(defined($rule->{after_fix})) &&
#				not(defined($rule->{before_fix})) &&
#				not(defined($rule->{token_fix})));
				
    my $left = $rule->{autofix}{after};
    my $right = $rule->{autofix}{before};
    
    my @subs = split /$tok/, $str;
    my $subs2 = join '#', @subs;
    debug("splitted string: $subs2");

    next if( scalar(@subs) == 1 );
    debug("fix_string: test left `$left' & right `$right'");
    
    for( my $i = 0; $i < scalar(@subs)-1 ; $i++ ){
      
      foreach my $validator ( @{ $rule->{validators} } ){
	debug("loop1");
	if( defined($validator->{after}) ){
	  my $l = $validator->{after};
	  debug("-->after: $l  ");
	  $valid = ($subs[$i] =~ /$l$/);
	  next if $valid == 0;
	}
	if( defined($validator->{before}) ){
	  my $r = $validator->{before};
	  debug("-->before: $r  ");
	  $valid = ($subs[$i+1] =~ /^$r/);
	}
	last if( $valid == 1 )
      }
      
      if( $valid == 0 ){
	debug("fix_string: invalid -> to be fixed");
	$subs[$i] = fix_right( $subs[$i], $left );
	$subs[$i+1] = fix_left( $subs[$i+1], $right );
	        }else{
		  debug("fix_string: valid -> skipping to next substring");	
	        }
    }
    $tok = $rule->{autofix}{token} if(defined($rule->{autofix}{token}));
    $str = join $tok, @subs;
    debug("intermediate fixed string: `$str'");
  }
  debug("fixed string: `$str'");
  return $str;
}

sub fix_left($$){
  my ($str, $left) = @_;
  
  if( $str =~ /^$left/ ){
    debug "fix_left: OK";
    return $str;
  }else{
    debug "fix_left: KO";
    return ($left .= $str);
  }
}

sub fix_right($$){
  my ($str, $right) = @_;
  
  if( $str =~ /$right$/ ){
    debug "fix_right: OK";
        return $str;
  }else{
    debug "fix_right: KO";
    return ($str .= $right);
  }
}

sub init($){
	my $mode = shift;

	debug "mode is set to `$mode'. Loading according rulesets...";

	load_rulesets( $mode );
}



BEGIN{
}

1;


