#!/usr/bin/perl

package PO_Reader;

use strict;
use diagnostics;

use File::Basename;

my $LOOK_AHEAD = '';
my $DEBUG = 0;
my $LINE = 0;
my $EXTERNAL_PO_FNAME = '';
my $INTERNAL_PO_FNAME = '';
my $INTERNAL_PO_DIR = '/tmp';
my $LOCAL_CHARSET = 'iso-8859-15';

sub _warn{
	while( defined( my $l = shift ) ){
		$l =~ s/\s*$//; # remove trailning spaces
		print STDERR "warning in $EXTERNAL_PO_FNAME: $l\n";
	}	
}

sub _debug{
	if( $DEBUG ){ 
		my ($c_package, $c_fname, $c_line) = caller( 1 );   
		while( defined( my $l = shift ) ){
			$l =~ s/\s*$//; # remove trailning spaces
			print STDERR "(debug at $c_fname:$c_line) $l\n";
		}
	}
}

sub _debug_enter_proc(){
	my ($c_package, $c_fname, $c_line, $func_name) = caller( 1 );   

	_debug "entered in function $func_name";
}

sub _abort{
	_debug_enter_proc();

	my ($c_package, $c_fname, $c_line) = caller(1);   
	while( defined( my $l = shift ) ){
		$l =~ s/\s*$//; # remove trailning spaces
		print STDERR "(abort at $c_fname:$c_line) $l\n";
	}
	print STDERR "Aborting...\n";

	po_close();
}

sub set_debug($){
	$DEBUG = shift;
}

sub _get_charset(){
	my $charset = '';
	
	_debug_enter_proc();

	open PO, $EXTERNAL_PO_FNAME or die "Can't open $EXTERNAL_PO_FNAME for read.";
	while( defined( my $l = <PO> ) ){
		  if( $l =~ m#charset=([a-zA-Z\-0-9]+)#g ){
			  $charset = $1;
			  last;
		  }
	}
	close PO;
	return $charset;
}

sub _open_internal_file(){
	_debug_enter_proc();

	_abort "Can't write in `$INTERNAL_PO_DIR'."
	  if( ! -w $INTERNAL_PO_DIR );

	my $ext_bname = basename $EXTERNAL_PO_FNAME;

	$INTERNAL_PO_FNAME = "$INTERNAL_PO_DIR/$ext_bname.$LOCAL_CHARSET";
	system "/bin/cp -f $EXTERNAL_PO_FNAME $INTERNAL_PO_FNAME";

	print STDERR "Recoding temp file to $LOCAL_CHARSET...\n";
	system "recode -t utf-8..$LOCAL_CHARSET $INTERNAL_PO_FNAME";

	open PO, $INTERNAL_PO_FNAME or _abort "Can't open $INTERNAL_PO_FNAME for read.";
}

sub po_open($){
	po_close()
	  if( $EXTERNAL_PO_FNAME ne '' );

	$EXTERNAL_PO_FNAME = shift;

	_debug_enter_proc();

	_debug "EXTERNAL_PO_FNAME = `$EXTERNAL_PO_FNAME'";

	# check the charset
	my $charset = _get_charset();

	# if charset is utf8, open a temporary file for working purpose
	if( ( $charset =~ /utf-8/i ) || ( $charset =~ /utf8/i ) ){
		# recode to local charset before using the file contents
		_open_internal_file();

	}elsif( $charset =~ /iso-8859-/i ){
		# warn the user. KDE doesn't use local encoded files anymore
		_warn "non-utf8 charset is used (`$charset').";
		$INTERNAL_PO_FNAME = $EXTERNAL_PO_FNAME;
		
		open PO, $INTERNAL_PO_FNAME or _abort "Can't open $INTERNAL_PO_FNAME for read.";

	}else{
		_abort "unknown charset `$charset'.";
	}
}

sub po_close(){
	_debug "about to close $EXTERNAL_PO_FNAME";

	_debug_enter_proc();

	$LOOK_AHEAD = '';
	$LINE = 0;
	$EXTERNAL_PO_FNAME = '';
	$INTERNAL_PO_FNAME = '';

	close PO;
}

sub read_line(){
	_debug_enter_proc();

	$LINE++;
	$LOOK_AHEAD = <PO>;
	return $LOOK_AHEAD;
}

sub po_get_msgstr(){
	my $str = '';
	my $l = '';
	
	_debug_enter_proc();

	_abort "expected msgstr not found at line $LINE"
	  if( $LOOK_AHEAD !~ /^\s*msgstr\s+"(.*)"/ );
	
	#	print "trouve un msgstr (`$1')\n";
	$str = $1;	
	while( defined( my $l = read_line() ) ){
		return $str if($l !~ /^\s*"(.*)"/ );
		my $t = $1;
		if( ($str =~ /[a-zA-Z0-9]$/) && ($str !~ /\\n$/) && ($t =~ /^[a-zA-Z0-9]/) ){
		  print STDERR "\n($EXTERNAL_PO_FNAME:$LINE) previous msgstr line doesn't end with a whitespace\n";
		}
		$str .= $t;
	}
	return $str;
}

sub po_get_msgid{
	my $str = '';
	my $l = '';
	my $fuzzy;

	_debug_enter_proc();

	while( defined( my $l = read_line() ) ){
		if( $l =~ /^\s*msgstr/ ){
			if( $fuzzy ){
				$fuzzy = 0;
			}else{
				_abort "msgstr found when a msgid was expected at line $LINE"
			}
		}
		
		if ( $l =~ /^\#,.*fuzzy/) {
			$fuzzy = 1;
		}                                                                                                                             

		next if( $l !~ /^\s*msgid\s+"(.*)"/ );
		next if $fuzzy;
		#               print "trouve un msgid (`$1')\n";
		$str = $1;      
		while( defined( my $l = read_line() ) ){
			return $str if($l !~ /^\s*"(.*)"/ );
			$str .= $1;
		}
		return $str;
	}
	return ;
}

sub line(){
	return $LINE;
}

BEGIN{
}

1;

